
var sqrtf = Math.sqrt
var tan = Math.tan
var powf = Math.pow
var M_PI = Math.PI
var fabs = (a)=>a<0?-a:a
var min=(a,b)=>a<b?a:b
var max=(a,b)=>a>b?a:b

function vec3(x,y,z){
    this.x = x || 0;this.y = y || 0;this.z = z || 0
}

vec3.prototype.set=function(x,y,z){this.x=x;this.y=y;this.z=z; return this}
vec3.prototype.copy=function(v){this.x=v.x;this.y=v.y;this.z=v.z; return this}
vec3.prototype.dot=function(v){return this.x*v.x+this.y*v.y+this.z*v.z}
vec3.prototype.add=function(v){this.x+=v.x;this.y+=v.y;this.z+=v.z; return this}
vec3.prototype.sub=function(v){this.x-=v.x;this.y-=v.y;this.z-=v.z; return this}
vec3.prototype.fmul=function(v){this.x*=v;this.y*=v;this.z*=v; return this}
vec3.prototype.length=function(v){return sqrtf(this.x*this.x+this.y*this.y+this.z*this.z)}
vec3.prototype.normalize=function(v){var len = this.length();return this.fmul(1.0/len)}
vec3.prototype.multiplyScalar = vec3.prototype.fmul
function tmpv3(v){return v?new vec3().copy(v):new vec3()}


function vec4(x,y,z,w){this.x = x || 0;this.y = y || 0;this.z = z || 0;this.w = w || 0;}
vec4.prototype.copy=function(v){this.x=v.x;this.y=v.y;this.z=v.z;this.w=v.w; return this}
function tmpv4(v){return v?new vec4().copy(v):new vec4()}


function Light(position,intensity) {
    this.position = position || new vec3();
    this.intensity = intensity || 1;
};

function Material(refractive_index,albedo,diffuse_color,specular_exponent){
    this.refractive_index = refractive_index||1;
    this.albedo = albedo||new vec4(1,0,0,0);
    this.diffuse_color = diffuse_color||new vec3(0,0,0);
    this.specular_exponent = specular_exponent||1;
};

Material.prototype.copy = function(from){
    this.refractive_index = from.refractive_index;
    this.albedo.copy(from.albedo)
    this.diffuse_color.copy(from.diffuse_color)
    this.specular_exponent = from.specular_exponent
    return this;
}
function tmpMaterial(v){return v?new Material().copy(v):new Material()}




function startCaching(){}
function resetCache(){}
if(true){//This vec3 cache shaves 300 msec off the render time by reusing the same datastructures for each pixel traced
    function cache(type){
        this.base = 0;
        this.count = 0;
        this.cache = [];
        this.type = type;
        type.prototype.cache = this;
    }
    cache.prototype.reset=function(){
        this.count = this.base;
    }
    cache.prototype.start=function(){
        this.base = this.count;
    }
    cache.prototype.alloc=function(v){
        var r;
        if(this.count<this.cache.length){
            r = this.cache[this.count];
            if(v)r.copy(v);
        }else{
            this.cache.push(r=(v?new this.type().copy(v):new this.type()));
        }
        this.count++
        return r;
    }
    var vec3cache = new cache(vec3);
    var vec4cache = new cache(vec4);
    var matCache = new cache(Material);
    resetCache=()=>{vec3cache.reset();vec4cache.reset();matCache.reset()} 
    startCaching=()=>{vec3cache.start();vec4cache.start();matCache.start()}
    tmpv3 = vec3cache.alloc.bind(vec3cache);
    tmpv4 = vec4cache.alloc.bind(vec4cache);
    tmpMaterial = matCache.alloc.bind(matCache);
}



var tv30 = new vec3();
var tv32 = new vec3()

var background_color = new vec3(0.2, 0.7, 0.8)
var epsilon = 1e-3;

var tlightdir=new vec3();
var treflectdir=new vec3();

var tv3h = new vec3()
var tv3i = new vec3()

var emptyMaterial = new Material();


function Sphere(center,radius,material){
    this.center = center || new vec3();
    this.radius = radius || 1;
    this.material = material || new Material();
};

Sphere.prototype.ray_intersect = function(orig,dir) {
    var L = tv30.copy(this.center).sub(orig);
    var tca = L.dot(dir);
    var d2 = L.dot(L)-tca*tca;
    if (d2 > this.radius*this.radius) return -1;
    var thc = sqrtf(this.radius*this.radius - d2);
    t0 = tca - thc;
    var t1 = tca + thc;
    if (t0 < 0) t0 = t1;
    return t0;
}

function reflect(I,N) {
    return tmpv3(I).sub(tv32.copy(N).multiplyScalar(2*(I.dot(N))));
}

function refract(I, N, refractive_index) { // Snell's law
    var cosi = - max(-1, min(1, I.dot(N)));
    var etai = 1, etat = refractive_index;
    var n = tmpv3(N);
    if (cosi < 0) { // if the ray is inside the object, swap the indices and invert the normal to get the correct result
        cosi = -cosi;
        var swp = etai;
        etai = etat;
        etat = swp;
        n.multiplyScalar(-1);
    }
    var eta = etai / etat;
    var k = 1 - eta*eta*(1 - cosi*cosi);
    return k < 0 ? tmpv3().set(0,0,0) : tmpv3(I).multiplyScalar(eta).add(n.multiplyScalar(eta * cosi - sqrtf(k)));
}

function scene_intersect(orig,dir,spheres, hit,N,material) {
    var spheres_dist = Number.MAX_VALUE;
    for (var i=0; i < spheres.length; i++) {
        var dist_i;
        dist_i=spheres[i].ray_intersect(orig, dir)
        if ((dist_i>=0) && (dist_i < spheres_dist)) {
            spheres_dist = dist_i;
            hit.copy(dir).multiplyScalar(dist_i).add(orig);
            N.copy(hit).sub(spheres[i].center).normalize();
            material.copy(spheres[i].material);
        }
    }

    var checkerboard_dist = Number.MAX_VALUE;
    if (fabs(dir.y)>epsilon)  {
        var d = -(orig.y+4)/dir.y; // the checkerboard plane has equation y = -4
        var pt = tmpv3(dir).multiplyScalar(d).add(orig);
        if (d>0 && fabs(pt.x)<10 && pt.z<-10 && pt.z>-30 && d<spheres_dist) {
            checkerboard_dist = d;
            hit.copy(pt);
            N.set(0,1,0);
            if((0|(.5*hit.x+1000)) + (0|(.5*hit.z)) & 1)
                material.diffuse_color.set(1,1,1);
            else 
                material.diffuse_color.set(1, .7, .3);
            material.diffuse_color.multiplyScalar(.3);
        }
    }
    var nearestDist = min(spheres_dist, checkerboard_dist) 
    return nearestDist < 1000;
}

function cast_ray(orig,dir, spheres, lights, depth=0) { //returns vec3
    var point=tmpv3();
    var N=tmpv3();
    var material = tmpMaterial(emptyMaterial);

    if ((depth>4) || (!scene_intersect(orig, dir, spheres, point, N, material))) {
        return background_color; // background color
    }

    var reflect_dir = tmpv3(reflect(dir, N)).normalize();
    var refract_dir = tmpv3(refract(dir, N, material.refractive_index)).normalize();
    var n_eps = tmpv3(N).multiplyScalar(epsilon)

    var reflect_orig = tmpv3(reflect_dir.dot(N) < 0 ? tmpv3(point).sub(n_eps) : tmpv3(point).add(n_eps)); // offset the original point to avoid occlusion by the object itself
    var refract_orig = tmpv3(refract_dir.dot(N) < 0 ? tmpv3(point).sub(n_eps) : tmpv3(point).add(n_eps));
    var reflect_color = cast_ray(reflect_orig, reflect_dir, spheres, lights, depth + 1);
    var refract_color = cast_ray(refract_orig, refract_dir, spheres, lights, depth + 1);

    var diffuse_light_intensity = 0, specular_light_intensity = 0;
    
    for (var i=0; i<lights.length; i++) {
        var light_dir      = tmpv3(lights[i].position).sub(point).normalize();
        var light_distance = tmpv3(lights[i].position).sub(point).length();

        var shadow_orig = light_dir.dot(N) < 0 ? tmpv3(point).sub(n_eps) : tmpv3(point).add(n_eps); // checking if the point lies in the shadow of the lights[i]
        var shadow_pt=tmpv3(), shadow_N=tmpv3();
        var tmpmaterial = tmpMaterial(emptyMaterial);
        if (scene_intersect(shadow_orig, light_dir, spheres, shadow_pt, shadow_N, tmpmaterial) && tmpv3(shadow_pt).sub(shadow_orig).length() < light_distance)
            continue;


        diffuse_light_intensity  += lights[i].intensity * max(0, light_dir.dot(N));


        var ilight = tlightdir.copy(light_dir).multiplyScalar(-1);
        var ref = reflect(ilight, N).multiplyScalar(-1)
        var refdot = treflectdir.copy(ref).dot(dir)
        specular_light_intensity += powf(max(0, refdot), material.specular_exponent)*lights[i].intensity;
        //specular_light_intensity += powf(std::max(0.f, -reflect(-light_dir, N)*dir), material.specular_exponent)*lights[i].intensity;
    }

    var tdiffuse = tmpv3(material.diffuse_color).multiplyScalar(diffuse_light_intensity * material.albedo.x)
    var tspecular = tmpv3().set(1., 1., 1.).multiplyScalar(specular_light_intensity * material.albedo.y)
    var treflect = tmpv3(reflect_color).multiplyScalar(material.albedo.z)
    var trefract = tmpv3(refract_color).multiplyScalar(material.albedo.w);

    return tdiffuse.add(tspecular).add(treflect).add(trefract);
}

var f2b=(f)=>(max(min(f*255.99,255),0)|0)

function *render(spheres, lights) {

    var width    = 1024;
    var height   = 768;
    var fov      = (M_PI/2.)|0;
    var framebuffer = document.createElement('canvas');
    framebuffer.width = width;
    framebuffer.height = height;

    var saveButton = document.createElement('button')
    saveButton.innerText = 'Rendering...'
    document.body.appendChild(saveButton);
    document.body.appendChild(document.createElement('br'));

    document.body.appendChild(framebuffer);

    var ctx = framebuffer.getContext('2d');
    ctx.fillText(10,10,'Rendering...')
    var imgData = ctx.getImageData(0,0,width,height);
    var pixels = imgData.data;

var startRegister = 0;

    var renderStart = performance.now();
    console.time("render")
    
startCaching()
    for (var j = 0; j<height; j++) {
        for (var i = 0; i<width; i++) {

            var x =  (2*(i + 0.5)/width  - 1)*tan(fov/2.)*width/height;
            var y = -(2*(j + 0.5)/height - 1)*tan(fov/2.);
            var dir = tv3h.set(x, y, -1).normalize();
            var color = cast_ray(tv3i.set(0,0,0), dir, spheres, lights);

            var pxid = (i+(j*width))*4
            pixels[pxid]=f2b(color.x)
            pixels[pxid+1]=f2b(color.y)
            pixels[pxid+2]=f2b(color.z)
            pixels[pxid+3]=255
resetCache();
        }
        if(!(j&63))yield; //Draw 32 rows at a time before yielding
        saveButton.innerText='Row:'+j
        ctx.putImageData(imgData,0,0);
    }

    var renderEndTime = performance.now();
    var sp=document.createElement('span');
    sp.innerText = " Render:"+((renderEndTime-renderStart)|0);
    saveButton.after(sp)
    console.timeEnd("render");

    saveButton.innerText = 'Save...'

    var imgdata = framebuffer.toDataURL('image/png');
    var newdata = imgdata.replace(/^data:image\/png/,'data:application/octet-stream');

    saveButton.onclick=()=>{
        var downloadElem = document.createElement('a')
        downloadElem.setAttribute('download','out.png')
        downloadElem.setAttribute('href',newdata)
        downloadElem.style.display = 'none';
        document.body.appendChild(downloadElem);
        downloadElem.click();
        document.body.removeChild(downloadElem);
    }
    yield 'done.'
}


function main() {
    var ivory=new Material(1.0, new vec4(0.6,  0.3, 0.1, 0.0), new vec3(0.4, 0.4, 0.3),   50.);
    var glass=new Material(1.5,  new vec4(0.0,  0.5, 0.1, 0.8), new vec3(0.6, 0.7, 0.8),  125.);
    var red_rubber=new Material(1.0,  new vec4(0.9,  0.1, 0.0, 0.0), new vec3(0.3, 0.1, 0.1),   10.);
    var mirror=new Material(1.0,  new vec4(0.0, 10.0, 0.8, 0.0), new vec3(1.0, 1.0, 1.0), 1425.);

    var spheres=[];
    spheres.push(new Sphere(new vec3(-3,    0,   -16), 2,      ivory));
    spheres.push(new Sphere(new vec3(-1.0, -1.5, -12), 2,      glass));
    spheres.push(new Sphere(new vec3( 1.5, -0.5, -18), 3, red_rubber));
    spheres.push(new Sphere(new vec3( 7,    5,   -18), 4,     mirror));

    var lights= [];
    lights.push(new Light(new vec3(-20, 20,  20), 1.5));
    lights.push(new Light(new vec3( 30, 50, -25), 1.8));
    lights.push(new Light(new vec3( 30, 20,  30), 1.7));

    var pixelGenerator = render(spheres, lights);
    var renderTask = setInterval(function(){
        if(pixelGenerator.next().value){
            clearInterval(renderTask);
        }
    },0);

    return 0;
}

main()